//
//  NURLConnection.m
//  Created by Rodrigo on 5/4/10.
//  https://github.com/rigoneri/ObjC-NURLConnection
//  
//  Copyright (c) 2011, Rodrigo Neri <@rigoneri>
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
//  
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
//  
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
//

#import "NURLConnection.h"

#define kNURLConnectionTimeout 30

@interface NURLConnection (hidden)
- (void)performRequest:(NSMutableURLRequest *)request;
- (void)initCache;
- (void)clearCache;
- (NSData *)retrieveFromCache:(NSString *)file;
- (NSString *)keyForURL:(NSString*)URL;
@end

@implementation NURLConnection

@synthesize delegate, receivedData, cachePath, filePath, previousCachedData;

- (id)init {
    if((self = [super init])) {
        receivedData = [[NSMutableData alloc] initWithLength:0];
        previousCachedData = [[NSData alloc] init];
        isCachedGetRequest = FALSE;
        isGetRequestWithCache = FALSE;
        status = 0;
        [self initCache];
    }
    return self;
}

- (void)dealloc {
    [receivedData release];
    [previousCachedData release];
    [cachePath release];
    [filePath release];
    [super dealloc];
}

#pragma mark -
#pragma mark Public Methods

//
//  Posts some data to a given url. The current content-type is application/json, for now change if needed.
//  If the post is succesfull it will invoke requestDidFinish and if not it will invoke requestDidFail:withError
//
- (void)postToURL:(NSString *)url withBody:(NSString *)body {
    [self postToURL:url withBody:body andToken:nil];
}

//
//  Posts some data to a given url. A token can be given for authorization in case needed
//  to call the service. The current content-type is application/json, for now change if needed.
//  If the post is succesfull it will invoke requestDidFinish and if not it will invoke requestDidFail:withError
//
- (void)postToURL:(NSString *)url withBody:(NSString *)body andToken:(NSString *)token {
    NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:url]
                                                           cachePolicy:NSURLRequestReloadIgnoringLocalCacheData 
                                                       timeoutInterval:kNURLConnectionTimeout];
	
    [request setHTTPMethod:@"POST"];
    [request setValue:@"application/json" forHTTPHeaderField:@"Content-type"];
    [request setHTTPBody:[body dataUsingEncoding:NSUTF8StringEncoding]];
    if(token)
        [request setValue:token forHTTPHeaderField:@"Authorization"];
    [self performRequest:request];
}

//
//  Gets some data from a given url. The current content-type is application/json, for now change if needed.
//  If the post is succesfull it will invoke requestDidFinish and if not it will invoke requestDidFail:withError
//
- (void)getFromURL:(NSString *)url {
	[self getFromURL:url andToken:nil];
}

//
//  Gets some data from a given url. A token can be given for authorization in case needed
//  to call the service. The current content-type is application/json, for now change if needed.
//  If the post is succesfull it will invoke requestDidFinish and if not it will invoke requestDidFail:withError
//
- (void)getFromURL:(NSString *)url andToken:(NSString *)token {
    NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:url]
                                                           cachePolicy:NSURLRequestReloadIgnoringLocalCacheData 
                                                       timeoutInterval:kNURLConnectionTimeout];
	
    [request setHTTPMethod:@"GET"];
    [request setValue:@"application/json" forHTTPHeaderField:@"Content-type"];
    if(token)
        [request setValue:token forHTTPHeaderField:@"Authorization"];
    [self performRequest:request];
}

// 
//  Gets some data from a given url with a modified since date. If the post is succesfull it will invoke 
//  requestDidFinish and if not it will invoke requestDidFail:withError
//	
- (void)getFromURL:(NSString *)url ifModifiedSince:(NSDate *)date {
	
    NSDateFormatter *formatter = [[NSDateFormatter alloc] init];
    NSLocale *usLocale = [[[NSLocale alloc] initWithLocaleIdentifier:@"en_US"] autorelease];
    [formatter setLocale: usLocale];
    [formatter setTimeZone: [NSTimeZone timeZoneWithName:@"GMT"]];
    [formatter setDateFormat: @"EEE, d MMM yyyy HH:mm:ss"];
    NSString *lastModified = [[formatter stringFromDate:date] stringByAppendingString:@" GMT"];
    [formatter release];
    
    NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:url]
                                                           cachePolicy:NSURLRequestReloadIgnoringLocalCacheData 
                                                       timeoutInterval:kNURLConnectionTimeout];
    [request setHTTPMethod:@"GET"];
    [request addValue:lastModified forHTTPHeaderField:@"If-Modified-Since"];
    [self performRequest:request];
}

// 
//  Gets some data from a given url or local cache. If the data is not found in cache it will make a request
//  to the given url and cache the response. An option to skip the cache can be given if needed. If the post 
//  is succesfull it will invoke requestDidFinish and if not it will invoke requestDidFail:withError
//	
- (void)getFromCacheOrURL:(NSString *)url skipCache:(BOOL)skipCache {
    
    filePath = [[cachePath stringByAppendingPathComponent:[self keyForURL:url]] retain];
    
    NSMutableData *cachedData = (NSMutableData *) [self retrieveFromCache:filePath];
    
    if (skipCache) {
        if(cachedData != nil) {
            //backing up the data in case of a connection error
            [previousCachedData release];
            previousCachedData = [[NSMutableData alloc] initWithData:cachedData]; 
            isCachedGetRequest = TRUE;
        }
        
        [self clearCache];
        cachedData = nil;
    }
    
    if(cachedData) {
        [receivedData release];
        receivedData = [[NSMutableData alloc] initWithData:cachedData];
        [[self delegate] requestDidFinish:self];
    }
    else {
        isGetRequestWithCache = TRUE;
        NSMutableURLRequest *request = [NSMutableURLRequest requestWithURL:[NSURL URLWithString:url]
                                                               cachePolicy:NSURLRequestReloadIgnoringLocalCacheData 
                                                           timeoutInterval:kNURLConnectionTimeout];
        [request setHTTPMethod:@"GET"];
		[request setValue:@"application/json" forHTTPHeaderField:@"Content-type"];
		[self performRequest:request];
    }
}

//
//  Returns the current connection status. If no connection will return 0. If an error will return the error code.
//
- (int)status {
    return status;
}

#pragma mark -
#pragma mark Private Methods

- (void)performRequest:(NSMutableURLRequest *)request {
    [receivedData release];
    receivedData = [[NSMutableData alloc] initWithLength:0];
    connection = [[[NSURLConnection alloc] initWithRequest:request 
                                                  delegate:self 
                                          startImmediately:YES] autorelease];
    
    if (connection == nil) {
        [[self delegate] requestDidFinish:self];
    }
}

#pragma mark -
#pragma mark NSURLConnection Delegate Methods

- (void)connection:(NSURLConnection *)connection didReceiveResponse:(NSURLResponse *)response {    
    if ([response respondsToSelector:@selector(statusCode)])
        status = [((NSHTTPURLResponse *) response) statusCode];
	
    [receivedData setLength:0];
}

- (void)connection:(NSURLConnection *)connection didReceiveData:(NSData *)data {
    [receivedData appendData:data];
}

- (void)connection:(NSURLConnection *)connection didFailWithError:(NSError *)error {
    if (isCachedGetRequest && previousCachedData) {
	    if ([[NSFileManager defaultManager] fileExistsAtPath:filePath] == NO) {
            [[NSFileManager defaultManager] createFileAtPath:filePath 
                                                    contents:previousCachedData
                                                  attributes:nil];
            
            [receivedData release];
            receivedData = [[NSMutableData alloc] initWithData:previousCachedData];
            [[self delegate] requestDidFinish:self];
        }
    }
    else {
	    status = error.code;
	    if (self.delegate != nil) {
            [[self delegate] requestDidFail:self withError:error];
	    }
    }
}

- (void)connectionDidFinishLoading:(NSURLConnection *)connection {
    if (isGetRequestWithCache && self.receivedData) {
        if ([[NSFileManager defaultManager] fileExistsAtPath:filePath] == NO) {
            [[NSFileManager defaultManager] createFileAtPath:filePath 
                                                    contents:self.receivedData
                                                  attributes:nil];
            
        }
    }
    
    if (self.delegate != nil)
        [[self delegate] requestDidFinish:self];
}

- (NSCachedURLResponse *)connection:(NSURLConnection *)connection willCacheResponse:(NSCachedURLResponse *)cachedResponse {
    return nil;
}

#pragma mark -
#pragma mark Cache

- (void)initCache {
    NSArray *paths = NSSearchPathForDirectoriesInDomains(NSCachesDirectory, NSUserDomainMask, YES);
    self.cachePath = [[paths objectAtIndex:0] stringByAppendingPathComponent:@"NURLCache"];
	
    if ([[NSFileManager defaultManager] fileExistsAtPath:cachePath]) 
        return;
	
    NSError *fileError;
    if (![[NSFileManager defaultManager] createDirectoryAtPath:cachePath
                                   withIntermediateDirectories:NO 
                                                    attributes:nil 
                                                         error:&fileError]) {
	    NSLog(@"Unable to create cache direcory: %@", [fileError localizedDescription]);
	    return;
    }
}

- (void)clearCache {
    if ([[NSFileManager defaultManager] fileExistsAtPath:filePath]) {
        NSError *fileError;
        if (![[NSFileManager defaultManager] removeItemAtPath:filePath error:&fileError]) {
            NSLog(@"Unable to clear cache: %@", [fileError localizedDescription]);
	    }
    }
}

- (NSData *)retrieveFromCache:(NSString *)file {
    if ([[NSFileManager defaultManager] fileExistsAtPath:file]) {
        NSData *cachedFile = [[NSData alloc] initWithContentsOfFile:file];
        return [cachedFile autorelease];
    }
    return nil;
}

#pragma mark -

- (NSString *)keyForURL:(NSString*)URL {
    const char* str = [URL UTF8String];
    unsigned char result[CC_MD5_DIGEST_LENGTH];
    CC_MD5(str, strlen(str), result);
	
    return [NSString stringWithFormat:
            @"%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
            result[0], result[1], result[2], result[3], result[4], result[5], result[6], result[7],
            result[8], result[9], result[10], result[11], result[12], result[13], result[14], result[15]
            ];
}

@end
